# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from .classwise import ClassWisePredictor
from .cluster import ClusteredPredictor
from .split import SplitPredictor
from .weight import WeightedPredictor
